﻿#Requires -RunAsAdministrator
# ------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All Rights Reserved. Licensed under the MIT License. See License in the project root for license information.
# ------------------------------------------------------------------------------

# This tool can be run on a Windows PC to automate creation of a bootable USB
# recovery key to automate the steps recommended by CrowdStrike:
# https://www.crowdstrike.com/blog/statement-on-falcon-content-update-for-windows-hosts/
# or to enable the device to boot in safe mode

# Constant Paths
$ADKInstallLocation = [System.Environment]::ExpandEnvironmentVariables("%ProgramFiles(x86)%\Windows Kits\10")
$ADKInstaller = [System.Environment]::ExpandEnvironmentVariables("%TEMP%\ADKSetup.exe")

$ADKWinPELocation = [System.Environment]::ExpandEnvironmentVariables("%ProgramFiles(x86)%\Windows Kits\10\Assessment and Deployment Kit\Windows Preinstallation Environment\amd64\en-us\winpe.wim")
$ADKWinPEMediaLocation = [System.Environment]::ExpandEnvironmentVariables("%ProgramFiles(x86)%\Windows Kits\10\Assessment and Deployment Kit\Windows Preinstallation Environment\amd64\Media")
$ADKWinPEAddOnInstaller = [System.Environment]::ExpandEnvironmentVariables("%TEMP%\adkwinpesetup.exe")

$WorkingLocation = [System.Environment]::ExpandEnvironmentVariables("%TEMP%\WinPEMountLocation\")
$WorkingWinPELocation = [System.Environment]::ExpandEnvironmentVariables("%ProgramFiles(x86)%\Windows Kits\10\Assessment and Deployment Kit\Windows Preinstallation Environment\amd64\en-us\winpe-bak.wim")
$CmdPromptPath = [System.Environment]::ExpandEnvironmentVariables("%WinDir%\System32\cmd.exe")
$WinPEMountLocation = [System.Environment]::ExpandEnvironmentVariables("%TEMP%\WinPEMountLocation\Mount")
$RecoveryImageLocation = [System.Environment]::ExpandEnvironmentVariables("%TEMP%\MsftRecoveryToolForCS.iso")

$DandISetEnvPath = [System.Environment]::ExpandEnvironmentVariables("%ProgramFiles(x86)%\Windows Kits\10\Assessment and Deployment Kit\Deployment Tools\DandISetEnv.bat")
$CopyPEPath = [System.Environment]::ExpandEnvironmentVariables("%ProgramFiles(x86)%\Windows Kits\10\Assessment and Deployment Kit\Windows Preinstallation Environment\copype.cmd")
$MakeWinPEMediaPath = [System.Environment]::ExpandEnvironmentVariables("%ProgramFiles(x86)%\Windows Kits\10\Assessment and Deployment Kit\Windows Preinstallation Environment\MakeWinPEMedia.cmd")

# Details
Write-Host "This tool will build a recovery image."

#
# Check if the ADK is installed
#
Write-Host "Checking if ADK is installed..."

$ADKInstalled = Test-Path -Path "$ADKInstallLocation\Assessment and Deployment Kit\Deployment Tools\amd64\Oscdimg"
if ($ADKInstalled)
{
    Write-Host "  -- An installation of ADK was found on device."
}
else
{
    Write-Host "  -- An installation of ADK was not found on the device."
    Write-Host "This tool will now download and install the Windows ADK."
    
    # Download the ADK Installer
    Write-Host "Downloading ADK..."
    
    # Remove existing installation file
    if (Test-Path $ADKInstaller)
    {
        Remove-Item $ADKInstaller -Verbose
    }
    
    # Download
    Invoke-WebRequest -Uri "https://go.microsoft.com/fwlink/?linkid=2271337" -OutFile $ADKInstaller
    
    # Verify hash
    if ((Get-FileHash $ADKInstaller).Hash -ne "3DBB9BF40E9CF5FACD9D770BE8EBA8F9509E77FC20A6051C0D9BAA1173F98E4B")
    {
        Write-Host "ERROR: Failed to verify ADK hash"
        Exit
    }

    $confirmation = Read-Host "Do you accept the Windows ADK license agreement (ADKLicenseAgreement.rtf) [Y]es or [N]o"
    if ($confirmation.ToUpperInvariant() -ne 'Y')
    {
        Exit
    }
    
    $confirmation = Read-Host "Allow Microsoft to collect insights for the Windows Kits as described in WindowsKitsPrivacy.rtf (optional) [Y]es or [N]o"
    $ceip = "/ceip off"

    if ($confirmation.ToUpperInvariant() -eq 'Y')
    {
        $ceip = "/ceip on"
    }

    Write-Host "Installing Windows ADK..."
    Start-Process -FilePath $ADKInstaller -ArgumentList '/features', '+', 'OptionId.DeploymentTools', '/q', $ceip -Wait
    Write-Host "  -- Successfully installed Windows ADK."
}

#
# Check if the ADK WinPE Addon is installed
#
Write-Host "Checking if ADK WinPE addon is installed..."
$ADKWinPEInstalled = Test-Path -Path $ADKWinPELocation
if ($ADKWinPEInstalled)
{
    Write-Host "  -- An installation of Windows ADK WinPE add-on was found on this device."
}
else
{
    Write-Host "  -- An installation for Windows ADK WinPE add-on was NOT found on this device."
    Write-Host "This tool will now download and install the Windows ADK WinPE add-on."
    
    # Download the Windows ADK WinPE add-on installer
    Write-Host "Downloading Windows ADK WinPE add-on..."
    
    # Remove existing installation file
    if (Test-Path $ADKWinPEAddOnInstaller)
    {
        Remove-Item $ADKWinPEAddOnInstaller -verbose
    }
    
    # Download
    Invoke-WebRequest -Uri "https://go.microsoft.com/fwlink/?linkid=2271338" -OutFile $ADKWinPEAddOnInstaller
    
    # Verify hash
    if ((Get-FileHash $ADKWinPEAddOnInstaller).Hash -ne "91AC010247B65244E5CD84C5F342D91B16501DBB08E422DE7DE06850CEF5680B")
    {
        Write-Host "ERROR: Failed to verify ADK WinPE add-on hash"
        Exit
    }
    
    # Confirm EULA
    $confirmation = Read-Host "Do you accept the Windows ADK license agreement (ADKLicenseAgreement.rtf) [Y]es or [N]o"
    if ($confirmation.ToUpperInvariant() -ne 'Y')
    {
        Exit
    }
    
    # Prompt for Ceip (if not already)
    if( $ceip.Length -eq 0 )
    {
        $confirmation = Read-Host "Allow Microsoft to collect insights for the Windows Kits as described in WindowsKitsPrivacy.rtf (optional) [Y]es or [N]o"
        $ceip = "/ceip off"

        if ($confirmation.ToUpperInvariant() -eq 'Y')
        {
            $ceip = "/ceip on"
        }
    }

    Write-Host "Installing the WinPE add-on..."
    Start-Process -FilePath $ADKWinPEAddOnInstaller -ArgumentList '/features', '+', 'OptionId.WindowsPreinstallationEnvironment', '/q', $ceip -Wait
    Write-Host "  -- Successfully installed Windows ADK WinPE add-on."
}

#
# Let admin pick safe boot or bitlocker key option before mounting the image
#
Write-Host ""
Write-Host "This script offers two options for recovering impacted devices:"
Write-Host "1. Boot to WinPE to remediate the issue. It requires entering bitlocker recovery key if system disk is bitlocker encrypted."
Write-Host "2. Boot to WinPE configure safe mode and run repair command after entering safe mode. This option is less likely to require bitlocker recovery key if system disk is bitlocker encrypted."
Write-Host ""
$winPEScriptOption = Read-Host "Which of the two options would you like to include in the WinPE image ? [1] or [2]"

# 
# Run the Deployment tools to set environment variables
#
Write-Host "Initializing Deployment Toolkit Environment..."

# Fetch the correct variables to be added.
$envVars = cmd.exe /c """$ADKInstallLocation\Assessment and Deployment Kit\Deployment Tools\DandISetEnv.bat"" && set" | Out-String

$envVars -split "`r`n" | ForEach-Object {
    if ($_ -match "^(.*?)=(.*)$")
    {
        # Update the current execution environment
        [System.Environment]::SetEnvironmentVariable($matches[1], $matches[2], [System.EnvironmentVariableTarget]::Process)
    }
}

#
# Use Dism to mount the WinPE image
#
Write-Host "Mounting WinPE image..."

if (!(Test-Path -Path $WinPEMountLocation))
{
    $mtDirRes = New-Item -Path $WinPEMountLocation -ItemType Directory
}
if (!(Test-Path -Path $WorkingWinPELocation))
{
    $wkDirRes = Copy-Item -Path "$ADKWinPELocation" -Destination "$WorkingWinPELocation" -Force
}

$mtResult = Mount-WindowsImage -ImagePath "$ADKWinPELocation" -Index 1 -Path "$WinPEMountLocation"

# Repair cmd file is located in the root folder of the media
$RepairCmdFile = "$ADKWinPEMediaLocation\repair.cmd"

# Remove any existing batch files
if (Test-Path "$WinPEMountLocation\CSRemediationScript.bat")
{
    Remove-Item "$WinPEMountLocation\CSRemediationScript.bat"
}

if (Test-Path "$RepairCmdFile")
{
    Remove-Item "$RepairCmdFile"
}

#
# Generate batch files based on the earlier selection of the recovery option
#
if ($winPEScriptOption.ToUpperInvariant() -eq '2')
{
    #
    # Generate batch file
    #
    "@echo off" | Out-File -FilePath "$WinPEMountLocation\CSRemediationScript.bat" -Append -Encoding ascii
    "echo This tool will configure this machine to boot in safe mode." | Out-File -FilePath "$WinPEMountLocation\CSRemediationScript.bat" -Append -Encoding ascii
    "echo WARNING: In some cases you may need to enter a BitLocker recovery key after running." | Out-File -FilePath "$WinPEMountLocation\CSRemediationScript.bat" -Force -Append -Encoding ascii
    "pause" | Out-File -FilePath "$WinPEMountLocation\CSRemediationScript.bat" -Force -Append -Encoding ascii
    "echo." | Out-File -FilePath "$WinPEMountLocation\CSRemediationScript.bat" -Force -Append -Encoding ascii
    "bcdedit /set {default} safeboot network" | Out-File -FilePath "$WinPEMountLocation\CSRemediationScript.bat" -Force -Append -Encoding ascii
    "echo." | Out-File -FilePath "$WinPEMountLocation\CSRemediationScript.bat" -Force -Append -Encoding ascii
    "IF %ERRORLEVEL% EQU 0 (" | Out-File -FilePath "$WinPEMountLocation\CSRemediationScript.bat" -Force -Append -Encoding ascii
    "     echo ................................................." | Out-File -FilePath "$WinPEMountLocation\CSRemediationScript.bat" -Force -Append -Encoding ascii
    "     echo Your PC is configured to boot to Safe Mode now.   " | Out-File -FilePath "$WinPEMountLocation\CSRemediationScript.bat" -Force -Append -Encoding ascii
    "     echo ................................................." | Out-File -FilePath "$WinPEMountLocation\CSRemediationScript.bat" -Force -Append -Encoding ascii
    "     echo If you manually changed the boot order on the device, restore the boot order to the previous state before rebooting. If BitLocker is enabled, make sure to remove the USB or bootable recovery device attached to prevent BitLocker recovery." | Out-File -FilePath "$WinPEMountLocation\CSRemediationScript.bat" -Force -Append -Encoding ascii
    "     echo ................................................." | Out-File -FilePath "$WinPEMountLocation\CSRemediationScript.bat" -Force -Append -Encoding ascii
    "echo." | Out-File -FilePath "$WinPEMountLocation\CSRemediationScript.bat" -Force -Append -Encoding ascii
    ") ELSE (" | Out-File -FilePath "$WinPEMountLocation\CSRemediationScript.bat" -Force -Append -Encoding ascii
    "     echo Could not configure safe mode on this system." | Out-File -FilePath "$WinPEMountLocation\CSRemediationScript.bat" -Force -Append -Encoding ascii
    "   )" | Out-File -FilePath "$WinPEMountLocation\CSRemediationScript.bat" -Force -Append -Encoding ascii
    "echo." | Out-File -FilePath "$WinPEMountLocation\CSRemediationScript.bat" -Force -Append -Encoding ascii
    "echo Upon reboot, run repair.cmd from the root of the media/USB drive to remove impacted files and disable safe mode." | Out-File -FilePath "$WinPEMountLocation\CSRemediationScript.bat" -Force -Append -Encoding ascii
    "echo." | Out-File -FilePath "$WinPEMountLocation\CSRemediationScript.bat" -Force -Append -Encoding ascii
    "pause" | Out-File -FilePath "$WinPEMountLocation\CSRemediationScript.bat" -Force -Append -Encoding ascii
    "exit 0" | Out-File -FilePath "$WinPEMountLocation\CSRemediationScript.bat" -Append -Encoding ascii

    #
    # Generate Repair.cmd
    #
    "@echo off" | Out-File -FilePath "$RepairCmdFile" -Append -Encoding ascii
    "echo This tool will remove impacted files and restore normal boot configuration." | Out-File -FilePath "$RepairCmdFile" -Append -Encoding ascii
    "echo." | Out-File -FilePath "$RepairCmdFile" -Append -Encoding ascii
    "echo WARNING: You may need BitLocker recovery key in some cases."  | Out-File -FilePath "$RepairCmdFile" -Append -Encoding ascii
    "echo WARNING: This script must be run in an elevated command prompt." | Out-File -FilePath "$RepairCmdFile" -Append -Encoding ascii
    "echo." | Out-File -FilePath "$RepairCmdFile" -Append -Encoding ascii
    "pause" | Out-File -FilePath "$RepairCmdFile" -Append -Encoding ascii
    "echo Removing impacted files..."  | Out-File -FilePath "$RepairCmdFile" -Append -Encoding ascii
    "del %SystemRoot%\System32\drivers\CrowdStrike\C-00000291*.sys" | Out-File -FilePath "$RepairCmdFile" -Append -Encoding ascii
    "echo Restoring normal boot flow..."  | Out-File -FilePath "$RepairCmdFile" -Append -Encoding ascii
    "bcdedit /deletevalue {current} safeboot" | Out-File -FilePath "$RepairCmdFile" -Append -Encoding ascii
    "echo Success. System will now reboot." | Out-File -FilePath "$RepairCmdFile" -Append -Encoding ascii
    "pause" | Out-File -FilePath "$RepairCmdFile" -Append -Encoding ascii
    "shutdown -r -t 00" | Out-File -FilePath "$RepairCmdFile" -Append -Encoding ascii
}
else 
{
    #
    # Generate batch file
    #
    "@echo on" | Out-File -FilePath "$WinPEMountLocation\CSRemediationScript.bat" -Force -Append -Encoding ascii
    "@setlocal enabledelayedexpansion" | Out-File -FilePath "$WinPEMountLocation\CSRemediationScript.bat" -Force -Append -Encoding ascii
    " " | Out-File -FilePath "$WinPEMountLocation\CSRemediationScript.bat" -Force -Append -Encoding ascii
    "@set drive=C:" | Out-File -FilePath "$WinPEMountLocation\CSRemediationScript.bat" -Force -Append -Encoding ascii
    ":tryagain" | Out-File -FilePath "$WinPEMountLocation\CSRemediationScript.bat" -Force -Append -Encoding ascii
    "@echo Using drive %drive%" | Out-File -FilePath "$WinPEMountLocation\CSRemediationScript.bat" -Force -Append -Encoding ascii
    "@echo." | Out-File -FilePath "$WinPEMountLocation\CSRemediationScript.bat" -Force -Append -Encoding ascii
    "@echo If your device is BitLocker encrypted use your phone to log on to https://aka.ms/aadrecoverykey. Log on with your Email ID and domain account password to find the BitLocker recovery key associated with your device." | Out-File -FilePath "$WinPEMountLocation\CSRemediationScript.bat" -Force -Append -Encoding ascii
    "@echo." | Out-File -FilePath "$WinPEMountLocation\CSRemediationScript.bat" -Force -Append -Encoding ascii
    " " | Out-File -FilePath "$WinPEMountLocation\CSRemediationScript.bat" -Force -Append -Encoding ascii
    "@rem If no key protectors are found, this command will return -1 and display ""ERROR: No key protectors found."". This error can be ignored." | Out-File -FilePath "$WinPEMountLocation\CSRemediationScript.bat" -Force -Append -Encoding ascii
    "manage-bde.exe -protectors %drive% -get -Type RecoveryPassword" | Out-File -FilePath "$WinPEMountLocation\CSRemediationScript.bat" -Force -Append -Encoding ascii
    " " | Out-File -FilePath "$WinPEMountLocation\CSRemediationScript.bat" -Force -Append -Encoding ascii
    "@IF NOT [%ERRORLEVEL%] == [0] @IF NOT [%ERRORLEVEL%] == [-1] (" | Out-File -FilePath "$WinPEMountLocation\CSRemediationScript.bat" -Force -Append -Encoding ascii
    "    @echo. ERROR: Failed with return code %ERRORLEVEL%" | Out-File -FilePath "$WinPEMountLocation\CSRemediationScript.bat" -Force -Append -Encoding ascii
    "    @echo." | Out-File -FilePath "$WinPEMountLocation\CSRemediationScript.bat" -Force -Append -Encoding ascii
    "    set /p continue=""Would you like to [C]ontinue or [T]ry another drive? [C/T] """ | Out-File -FilePath "$WinPEMountLocation\CSRemediationScript.bat" -Force -Append -Encoding ascii
    "    @if [!continue!] == [C] goto :next" | Out-File -FilePath "$WinPEMountLocation\CSRemediationScript.bat" -Force -Append -Encoding ascii
    " " | Out-File -FilePath "$WinPEMountLocation\CSRemediationScript.bat" -Force -Append -Encoding ascii
    "    @set /p drive=""Enter alternate drive letter to try (e.g. Z:): """ | Out-File -FilePath "$WinPEMountLocation\CSRemediationScript.bat" -Force -Append -Encoding ascii
    "    @goto :tryagain" | Out-File -FilePath "$WinPEMountLocation\CSRemediationScript.bat" -Force -Append -Encoding ascii
    ")" | Out-File -FilePath "$WinPEMountLocation\CSRemediationScript.bat" -Force -Append -Encoding ascii
    " " | Out-File -FilePath "$WinPEMountLocation\CSRemediationScript.bat" -Force -Append -Encoding ascii
    ":next" | Out-File -FilePath "$WinPEMountLocation\CSRemediationScript.bat" -Force -Append -Encoding ascii
    "@echo." | Out-File -FilePath "$WinPEMountLocation\CSRemediationScript.bat" -Force -Append -Encoding ascii
    " " | Out-File -FilePath "$WinPEMountLocation\CSRemediationScript.bat" -Force -Append -Encoding ascii
    "@set /p reckey=""Enter recovery key for this drive if required: """ | Out-File -FilePath "$WinPEMountLocation\CSRemediationScript.bat" -Force -Append -Encoding ascii
    "@IF NOT [%reckey%] == [] (" | Out-File -FilePath "$WinPEMountLocation\CSRemediationScript.bat" -Force -Append -Encoding ascii
    "    @echo Unlocking drive %drive%" | Out-File -FilePath "$WinPEMountLocation\CSRemediationScript.bat" -Force -Append -Encoding ascii
    "    manage-bde.exe -unlock %drive% -recoverypassword %reckey%" | Out-File -FilePath "$WinPEMountLocation\CSRemediationScript.bat" -Force -Append -Encoding ascii
    "    IF NOT [!ERRORLEVEL!] == [0] @IF NOT [!ERRORLEVEL!] == [-1] (" | Out-File -FilePath "$WinPEMountLocation\CSRemediationScript.bat" -Force -Append -Encoding ascii
    "        @echo. ERROR: Failed with return code %ERRORLEVEL%" | Out-File -FilePath "$WinPEMountLocation\CSRemediationScript.bat" -Force -Append -Encoding ascii
    "        @echo." | Out-File -FilePath "$WinPEMountLocation\CSRemediationScript.bat" -Force -Append -Encoding ascii
    "        set /p continue=""Would you like to [C]ontinue or [T]ry another drive? [C/T] """ | Out-File -FilePath "$WinPEMountLocation\CSRemediationScript.bat" -Force -Append -Encoding ascii
    "        @if [!continue!] == [C] goto :next2" | Out-File -FilePath "$WinPEMountLocation\CSRemediationScript.bat" -Force -Append -Encoding ascii
    " " | Out-File -FilePath "$WinPEMountLocation\CSRemediationScript.bat" -Force -Append -Encoding ascii
    "        @set /p drive=""Enter alternate drive letter to try (e.g. Z:): """ | Out-File -FilePath "$WinPEMountLocation\CSRemediationScript.bat" -Force -Append -Encoding ascii
    "        @goto :tryagain" | Out-File -FilePath "$WinPEMountLocation\CSRemediationScript.bat" -Force -Append -Encoding ascii
    "    )    " | Out-File -FilePath "$WinPEMountLocation\CSRemediationScript.bat" -Force -Append -Encoding ascii
    ")" | Out-File -FilePath "$WinPEMountLocation\CSRemediationScript.bat" -Force -Append -Encoding ascii
    ":next2" | Out-File -FilePath "$WinPEMountLocation\CSRemediationScript.bat" -Force -Append -Encoding ascii
    "@set targetfiles=%drive%\Windows\System32\drivers\CrowdStrike\C-00000291*.sys" | Out-File -FilePath "$WinPEMountLocation\CSRemediationScript.bat" -Force -Append -Encoding ascii
    "@if exist %targetfiles% (" | Out-File -FilePath "$WinPEMountLocation\CSRemediationScript.bat" -Force -Append -Encoding ascii
    "    @echo. Target file^(s^) %targetfiles% detected:" | Out-File -FilePath "$WinPEMountLocation\CSRemediationScript.bat" -Force -Append -Encoding ascii
    "    @echo." | Out-File -FilePath "$WinPEMountLocation\CSRemediationScript.bat" -Force -Append -Encoding ascii
    "    @dir %targetfiles%" | Out-File -FilePath "$WinPEMountLocation\CSRemediationScript.bat" -Force -Append -Encoding ascii
    "    @echo. Removing file^(s^)..." | Out-File -FilePath "$WinPEMountLocation\CSRemediationScript.bat" -Force -Append -Encoding ascii
    "    del /f %targetfiles%" | Out-File -FilePath "$WinPEMountLocation\CSRemediationScript.bat" -Force -Append -Encoding ascii
    "    @echo Done performing cleanup operation." | Out-File -FilePath "$WinPEMountLocation\CSRemediationScript.bat" -Force -Append -Encoding ascii
    ") else (" | Out-File -FilePath "$WinPEMountLocation\CSRemediationScript.bat" -Force -Append -Encoding ascii
    "    @echo Target files were not detected." | Out-File -FilePath "$WinPEMountLocation\CSRemediationScript.bat" -Force -Append -Encoding ascii
    ")" | Out-File -FilePath "$WinPEMountLocation\CSRemediationScript.bat" -Force -Append -Encoding ascii
    ":end" | Out-File -FilePath "$WinPEMountLocation\CSRemediationScript.bat" -Force -Append -Encoding ascii
    "@echo." | Out-File -FilePath "$WinPEMountLocation\CSRemediationScript.bat" -Force -Append -Encoding ascii
    "@pause" | Out-File -FilePath "$WinPEMountLocation\CSRemediationScript.bat" -Force -Append -Encoding ascii
    "@exit 0" | Out-File -FilePath "$WinPEMountLocation\CSRemediationScript.bat" -Force -Append -Encoding ascii
}

#
# Generate WinPEShl.ini file to autolaunch recovery script
#
"[LaunchApps]" | Out-File -FilePath "$WinPEMountLocation\Windows\system32\winpeshl.ini" -Force -Encoding ascii
"%SYSTEMDRIVE%\Windows\system32\cmd.exe /k %SYSTEMDRIVE%\CSRemediationScript.bat" | Out-File -FilePath "$WinPEMountLocation\Windows\system32\winpeshl.ini" -Append -Encoding ascii

# Add necessary packages
Write-Host "Adding necessary packages..."

# WinPE-WMI.cab
$pkgWmiResult = Add-WindowsPackage -Path "$WinPEMountLocation" -PackagePath "$ADKInstallLocation\Assessment and Deployment Kit\Windows Preinstallation Environment\amd64\WinPE_OCs\WinPE-WMI.cab"
$pkgWmiLngResult = Add-WindowsPackage -Path "$WinPEMountLocation" -PackagePath "$ADKInstallLocation\Assessment and Deployment Kit\Windows Preinstallation Environment\amd64\WinPE_OCs\en-us\WinPE-WMI_en-us.cab"

# WinPE-SecureStartup.cab
$pkgStartResult=Add-WindowsPackage -Path "$WinPEMountLocation" -PackagePath "$ADKInstallLocation\Assessment and Deployment Kit\Windows Preinstallation Environment\amd64\WinPE_OCs\WinPE-SecureStartup.cab"
$pkgStartLngResult = Add-WindowsPackage -Path "$WinPEMountLocation" -PackagePath "$ADKInstallLocation\Assessment and Deployment Kit\Windows Preinstallation Environment\amd64\WinPE_OCs\en-us\WinPE-SecureStartup_en-us.cab"

#
# Optionally add drivers to the WinPE image
#
$confirmation = Read-Host "Do you need to add drivers to the WinPE image ? [Y]es or [N]o"
if ($confirmation.ToUpperInvariant() -eq 'Y')
{
    $driverPath = Read-Host "Specify the folder that contains subfolders with driver (.ini) files or press Enter to skip"

    if ($driverPath -ne "")
    {
        Write-Host "Adding drivers..."
        Add-WindowsDriver -Path "$WinPEMountLocation" -Driver "$driverPath" -Recurse
    }
}

#
# Unmount and commit WinPE image
#
Write-Host "Saving the changes to the WinPE image..."
$disImgResult = Dismount-WindowsImage -Path "$WinPEMountLocation" -Save

#
# Creates working directories for WinPE media creation
#
$WinPEStagingLocation = [System.Environment]::ExpandEnvironmentVariables("%TEMP%\WinPEStagingLocation")

if ((Test-Path -Path $WinPEStagingLocation))
{
    $WinPeRemResult = Remove-Item -Path $WinPEStagingLocation -Force -Recurse
}

[System.Environment]::SetEnvironmentVariable("WinPERoot", "$ADKInstallLocation\Assessment and Deployment Kit\Windows Preinstallation Environment")
[System.Environment]::SetEnvironmentVariable("OSCDImgRoot", "$ADKInstallLocation\Assessment and Deployment Kit\Deployment Tools\AMD64\Oscdimg")

$cmdArgs = "amd64 " + "`"$WinPEStagingLocation`""
Start-Process -FilePath $CopyPEPath -ArgumentList $cmdArgs -Wait -NoNewWindow

Write-Host "Creating ISO..."

# Create the ISO
if (Test-Path -Path $RecoveryImageLocation)
{
    Remove-Item -Path $RecoveryImageLocation -Force
}

$CmdArgs = "/ISO " + "`"$WinPEStagingLocation`" `"$RecoveryImageLocation`""
Start-Process -FilePath $MakeWinPEMediaPath -ArgumentList $cmdArgs -Wait -NoNewWindow

#
# Prompt if iso or USB is needed
#
$isUsb = Read-Host "Do you need an ISO [1] or a USB [2] ?"
if ($isUsb.ToUpperInvariant() -eq '2')
{    
    #
    # Make USB Key
    #

    $USBDrive = Read-Host "What is the drive letter of your USB Key?"

    if ($USBDrive.Length -lt 1)
    {
        Write-Host "ERROR: Invalid drive letter"
        Exit
    }

    if ($USBDrive.Length -eq 1)
    {
        $USBDrive = -join ($USBDrive, ":")
    }

    if (!(Test-Path $USBDrive))
    {
        Write-Host "ERROR: Drive not found"
        Exit
    }

    $usbVolume = Get-Volume -DriveLetter $USBDrive[0]
    if (($usbVolume.Size) -gt 32GB)
    {
        Write-Host "ERROR: USB drives larger than 32GB are not supported. Please shrink the drive partitions and re-run the script."
        Exit
    }

    Format-Volume -DriveLetter $USBDrive[0] -FileSystem FAT32

    Write-Host "Making USB media..."

    # Mount the ISO
    $mountVolume = Mount-DiskImage -ImagePath "$RecoveryImageLocation" -PassThru
    $mountLetter = ($mountVolume | Get-Volume).DriveLetter + ":\*"

    Write-Host "Copying contents to the USB drive..."
    Copy-Item -Path $mountLetter -Destination "$USBDrive\" -Recurse

    Write-Host "Cleaning up..."
    $dismountResult = Dismount-DiskImage -ImagePath "$RecoveryImageLocation"

    Write-Host "DONE: You can now boot from the USB key."

}
else {
    Write-Host "ISO is available here: $RecoveryImageLocation"
}

if (Test-Path -Path $WinPEStagingLocation)
{
    $remStgResult = Remove-Item -Path $WinPEStagingLocation -Force -Recurse
}

if (($isUsb -eq $true) -and (Test-Path -Path $RecoveryImageLocation))
{
    $remImgResult = Remove-Item -Path $RecoveryImageLocation -Force
}

if (Test-Path -Path $WorkingLocation)
{
    $remWkResult = Remove-Item -Path $WorkingLocation -Force -Recurse
}



# SIG # Begin signature block
# MIIoSgYJKoZIhvcNAQcCoIIoOzCCKDcCAQExDzANBglghkgBZQMEAgEFADB5Bgor
# BgEEAYI3AgEEoGswaTA0BgorBgEEAYI3AgEeMCYCAwEAAAQQH8w7YFlLCE63JNLG
# KX7zUQIBAAIBAAIBAAIBAAIBADAxMA0GCWCGSAFlAwQCAQUABCBfs42Kl7qnXAr7
# jelVLxCYf2FdnoiJWXgfVFlLdzg7QKCCDXYwggX0MIID3KADAgECAhMzAAADrzBA
# DkyjTQVBAAAAAAOvMA0GCSqGSIb3DQEBCwUAMH4xCzAJBgNVBAYTAlVTMRMwEQYD
# VQQIEwpXYXNoaW5ndG9uMRAwDgYDVQQHEwdSZWRtb25kMR4wHAYDVQQKExVNaWNy
# b3NvZnQgQ29ycG9yYXRpb24xKDAmBgNVBAMTH01pY3Jvc29mdCBDb2RlIFNpZ25p
# bmcgUENBIDIwMTEwHhcNMjMxMTE2MTkwOTAwWhcNMjQxMTE0MTkwOTAwWjB0MQsw
# CQYDVQQGEwJVUzETMBEGA1UECBMKV2FzaGluZ3RvbjEQMA4GA1UEBxMHUmVkbW9u
# ZDEeMBwGA1UEChMVTWljcm9zb2Z0IENvcnBvcmF0aW9uMR4wHAYDVQQDExVNaWNy
# b3NvZnQgQ29ycG9yYXRpb24wggEiMA0GCSqGSIb3DQEBAQUAA4IBDwAwggEKAoIB
# AQDOS8s1ra6f0YGtg0OhEaQa/t3Q+q1MEHhWJhqQVuO5amYXQpy8MDPNoJYk+FWA
# hePP5LxwcSge5aen+f5Q6WNPd6EDxGzotvVpNi5ve0H97S3F7C/axDfKxyNh21MG
# 0W8Sb0vxi/vorcLHOL9i+t2D6yvvDzLlEefUCbQV/zGCBjXGlYJcUj6RAzXyeNAN
# xSpKXAGd7Fh+ocGHPPphcD9LQTOJgG7Y7aYztHqBLJiQQ4eAgZNU4ac6+8LnEGAL
# go1ydC5BJEuJQjYKbNTy959HrKSu7LO3Ws0w8jw6pYdC1IMpdTkk2puTgY2PDNzB
# tLM4evG7FYer3WX+8t1UMYNTAgMBAAGjggFzMIIBbzAfBgNVHSUEGDAWBgorBgEE
# AYI3TAgBBggrBgEFBQcDAzAdBgNVHQ4EFgQURxxxNPIEPGSO8kqz+bgCAQWGXsEw
# RQYDVR0RBD4wPKQ6MDgxHjAcBgNVBAsTFU1pY3Jvc29mdCBDb3Jwb3JhdGlvbjEW
# MBQGA1UEBRMNMjMwMDEyKzUwMTgyNjAfBgNVHSMEGDAWgBRIbmTlUAXTgqoXNzci
# tW2oynUClTBUBgNVHR8ETTBLMEmgR6BFhkNodHRwOi8vd3d3Lm1pY3Jvc29mdC5j
# b20vcGtpb3BzL2NybC9NaWNDb2RTaWdQQ0EyMDExXzIwMTEtMDctMDguY3JsMGEG
# CCsGAQUFBwEBBFUwUzBRBggrBgEFBQcwAoZFaHR0cDovL3d3dy5taWNyb3NvZnQu
# Y29tL3BraW9wcy9jZXJ0cy9NaWNDb2RTaWdQQ0EyMDExXzIwMTEtMDctMDguY3J0
# MAwGA1UdEwEB/wQCMAAwDQYJKoZIhvcNAQELBQADggIBAISxFt/zR2frTFPB45Yd
# mhZpB2nNJoOoi+qlgcTlnO4QwlYN1w/vYwbDy/oFJolD5r6FMJd0RGcgEM8q9TgQ
# 2OC7gQEmhweVJ7yuKJlQBH7P7Pg5RiqgV3cSonJ+OM4kFHbP3gPLiyzssSQdRuPY
# 1mIWoGg9i7Y4ZC8ST7WhpSyc0pns2XsUe1XsIjaUcGu7zd7gg97eCUiLRdVklPmp
# XobH9CEAWakRUGNICYN2AgjhRTC4j3KJfqMkU04R6Toyh4/Toswm1uoDcGr5laYn
# TfcX3u5WnJqJLhuPe8Uj9kGAOcyo0O1mNwDa+LhFEzB6CB32+wfJMumfr6degvLT
# e8x55urQLeTjimBQgS49BSUkhFN7ois3cZyNpnrMca5AZaC7pLI72vuqSsSlLalG
# OcZmPHZGYJqZ0BacN274OZ80Q8B11iNokns9Od348bMb5Z4fihxaBWebl8kWEi2O
# PvQImOAeq3nt7UWJBzJYLAGEpfasaA3ZQgIcEXdD+uwo6ymMzDY6UamFOfYqYWXk
# ntxDGu7ngD2ugKUuccYKJJRiiz+LAUcj90BVcSHRLQop9N8zoALr/1sJuwPrVAtx
# HNEgSW+AKBqIxYWM4Ev32l6agSUAezLMbq5f3d8x9qzT031jMDT+sUAoCw0M5wVt
# CUQcqINPuYjbS1WgJyZIiEkBMIIHejCCBWKgAwIBAgIKYQ6Q0gAAAAAAAzANBgkq
# hkiG9w0BAQsFADCBiDELMAkGA1UEBhMCVVMxEzARBgNVBAgTCldhc2hpbmd0b24x
# EDAOBgNVBAcTB1JlZG1vbmQxHjAcBgNVBAoTFU1pY3Jvc29mdCBDb3Jwb3JhdGlv
# bjEyMDAGA1UEAxMpTWljcm9zb2Z0IFJvb3QgQ2VydGlmaWNhdGUgQXV0aG9yaXR5
# IDIwMTEwHhcNMTEwNzA4MjA1OTA5WhcNMjYwNzA4MjEwOTA5WjB+MQswCQYDVQQG
# EwJVUzETMBEGA1UECBMKV2FzaGluZ3RvbjEQMA4GA1UEBxMHUmVkbW9uZDEeMBwG
# A1UEChMVTWljcm9zb2Z0IENvcnBvcmF0aW9uMSgwJgYDVQQDEx9NaWNyb3NvZnQg
# Q29kZSBTaWduaW5nIFBDQSAyMDExMIICIjANBgkqhkiG9w0BAQEFAAOCAg8AMIIC
# CgKCAgEAq/D6chAcLq3YbqqCEE00uvK2WCGfQhsqa+laUKq4BjgaBEm6f8MMHt03
# a8YS2AvwOMKZBrDIOdUBFDFC04kNeWSHfpRgJGyvnkmc6Whe0t+bU7IKLMOv2akr
# rnoJr9eWWcpgGgXpZnboMlImEi/nqwhQz7NEt13YxC4Ddato88tt8zpcoRb0Rrrg
# OGSsbmQ1eKagYw8t00CT+OPeBw3VXHmlSSnnDb6gE3e+lD3v++MrWhAfTVYoonpy
# 4BI6t0le2O3tQ5GD2Xuye4Yb2T6xjF3oiU+EGvKhL1nkkDstrjNYxbc+/jLTswM9
# sbKvkjh+0p2ALPVOVpEhNSXDOW5kf1O6nA+tGSOEy/S6A4aN91/w0FK/jJSHvMAh
# dCVfGCi2zCcoOCWYOUo2z3yxkq4cI6epZuxhH2rhKEmdX4jiJV3TIUs+UsS1Vz8k
# A/DRelsv1SPjcF0PUUZ3s/gA4bysAoJf28AVs70b1FVL5zmhD+kjSbwYuER8ReTB
# w3J64HLnJN+/RpnF78IcV9uDjexNSTCnq47f7Fufr/zdsGbiwZeBe+3W7UvnSSmn
# Eyimp31ngOaKYnhfsi+E11ecXL93KCjx7W3DKI8sj0A3T8HhhUSJxAlMxdSlQy90
# lfdu+HggWCwTXWCVmj5PM4TasIgX3p5O9JawvEagbJjS4NaIjAsCAwEAAaOCAe0w
# ggHpMBAGCSsGAQQBgjcVAQQDAgEAMB0GA1UdDgQWBBRIbmTlUAXTgqoXNzcitW2o
# ynUClTAZBgkrBgEEAYI3FAIEDB4KAFMAdQBiAEMAQTALBgNVHQ8EBAMCAYYwDwYD
# VR0TAQH/BAUwAwEB/zAfBgNVHSMEGDAWgBRyLToCMZBDuRQFTuHqp8cx0SOJNDBa
# BgNVHR8EUzBRME+gTaBLhklodHRwOi8vY3JsLm1pY3Jvc29mdC5jb20vcGtpL2Ny
# bC9wcm9kdWN0cy9NaWNSb29DZXJBdXQyMDExXzIwMTFfMDNfMjIuY3JsMF4GCCsG
# AQUFBwEBBFIwUDBOBggrBgEFBQcwAoZCaHR0cDovL3d3dy5taWNyb3NvZnQuY29t
# L3BraS9jZXJ0cy9NaWNSb29DZXJBdXQyMDExXzIwMTFfMDNfMjIuY3J0MIGfBgNV
# HSAEgZcwgZQwgZEGCSsGAQQBgjcuAzCBgzA/BggrBgEFBQcCARYzaHR0cDovL3d3
# dy5taWNyb3NvZnQuY29tL3BraW9wcy9kb2NzL3ByaW1hcnljcHMuaHRtMEAGCCsG
# AQUFBwICMDQeMiAdAEwAZQBnAGEAbABfAHAAbwBsAGkAYwB5AF8AcwB0AGEAdABl
# AG0AZQBuAHQALiAdMA0GCSqGSIb3DQEBCwUAA4ICAQBn8oalmOBUeRou09h0ZyKb
# C5YR4WOSmUKWfdJ5DJDBZV8uLD74w3LRbYP+vj/oCso7v0epo/Np22O/IjWll11l
# hJB9i0ZQVdgMknzSGksc8zxCi1LQsP1r4z4HLimb5j0bpdS1HXeUOeLpZMlEPXh6
# I/MTfaaQdION9MsmAkYqwooQu6SpBQyb7Wj6aC6VoCo/KmtYSWMfCWluWpiW5IP0
# wI/zRive/DvQvTXvbiWu5a8n7dDd8w6vmSiXmE0OPQvyCInWH8MyGOLwxS3OW560
# STkKxgrCxq2u5bLZ2xWIUUVYODJxJxp/sfQn+N4sOiBpmLJZiWhub6e3dMNABQam
# ASooPoI/E01mC8CzTfXhj38cbxV9Rad25UAqZaPDXVJihsMdYzaXht/a8/jyFqGa
# J+HNpZfQ7l1jQeNbB5yHPgZ3BtEGsXUfFL5hYbXw3MYbBL7fQccOKO7eZS/sl/ah
# XJbYANahRr1Z85elCUtIEJmAH9AAKcWxm6U/RXceNcbSoqKfenoi+kiVH6v7RyOA
# 9Z74v2u3S5fi63V4GuzqN5l5GEv/1rMjaHXmr/r8i+sLgOppO6/8MO0ETI7f33Vt
# Y5E90Z1WTk+/gFcioXgRMiF670EKsT/7qMykXcGhiJtXcVZOSEXAQsmbdlsKgEhr
# /Xmfwb1tbWrJUnMTDXpQzTGCGiowghomAgEBMIGVMH4xCzAJBgNVBAYTAlVTMRMw
# EQYDVQQIEwpXYXNoaW5ndG9uMRAwDgYDVQQHEwdSZWRtb25kMR4wHAYDVQQKExVN
# aWNyb3NvZnQgQ29ycG9yYXRpb24xKDAmBgNVBAMTH01pY3Jvc29mdCBDb2RlIFNp
# Z25pbmcgUENBIDIwMTECEzMAAAOvMEAOTKNNBUEAAAAAA68wDQYJYIZIAWUDBAIB
# BQCggc4wGQYJKoZIhvcNAQkDMQwGCisGAQQBgjcCAQQwHAYKKwYBBAGCNwIBCzEO
# MAwGCisGAQQBgjcCARUwLwYJKoZIhvcNAQkEMSIEIAoXSDOkhDHKCV2iOyb29vc8
# 7dV/B1c0exEFQIZwQKkSMGIGCisGAQQBgjcCAQwxVDBSoDSAMgBNAHMAZgB0AFIA
# ZQBjAG8AdgBlAHIAeQBUAG8AbwBsAEYAbwByAEMAUwAuAHAAcwAxoRqAGGh0dHA6
# Ly93d3cubWljcm9zb2Z0LmNvbTANBgkqhkiG9w0BAQEFAASCAQB8gzgcmlO+XIv0
# euaS5UUc9745fXOUElTGW285znSm83bGAA9GDH+yWTZyoYj/BOFRpTqF8RP/RFS7
# pvKAj17NSYLrvDXPVxKNb/INE9TGD4hbyENDPnjKIOweTGC87di+Bt16OEhGSCRJ
# bPfNFHr4K+2FdU4EMG8+XW1YBYTy5TlQPu5RnjbYirC8mfGeXJeZIlEo1Ci+ePs/
# zIBSP4ZoYUY/Q9po9lR8bEOtv1wqTpn0QLq4GokuKTYfjl2wdgV+7QtJmw08AsCj
# hSYXMqQFWwJ1K0NL9SDRk4jdhwJnfehZWH4xRXnDgNG/7y79Fg52SFd4vhBqiGUY
# 6li6P+MCoYIXlDCCF5AGCisGAQQBgjcDAwExgheAMIIXfAYJKoZIhvcNAQcCoIIX
# bTCCF2kCAQMxDzANBglghkgBZQMEAgEFADCCAVIGCyqGSIb3DQEJEAEEoIIBQQSC
# AT0wggE5AgEBBgorBgEEAYRZCgMBMDEwDQYJYIZIAWUDBAIBBQAEIFUHwzzdvftI
# 0L9oFdgfpjO4jkoPTydPwM1YeM8XlwvLAgZmlV4UkfYYEzIwMjQwNzIzMDE0ODM1
# LjA5MlowBIACAfSggdGkgc4wgcsxCzAJBgNVBAYTAlVTMRMwEQYDVQQIEwpXYXNo
# aW5ndG9uMRAwDgYDVQQHEwdSZWRtb25kMR4wHAYDVQQKExVNaWNyb3NvZnQgQ29y
# cG9yYXRpb24xJTAjBgNVBAsTHE1pY3Jvc29mdCBBbWVyaWNhIE9wZXJhdGlvbnMx
# JzAlBgNVBAsTHm5TaGllbGQgVFNTIEVTTjpBMDAwLTA1RTAtRDk0NzElMCMGA1UE
# AxMcTWljcm9zb2Z0IFRpbWUtU3RhbXAgU2VydmljZaCCEeowggcgMIIFCKADAgEC
# AhMzAAAB6+AYbLW27zjtAAEAAAHrMA0GCSqGSIb3DQEBCwUAMHwxCzAJBgNVBAYT
# AlVTMRMwEQYDVQQIEwpXYXNoaW5ndG9uMRAwDgYDVQQHEwdSZWRtb25kMR4wHAYD
# VQQKExVNaWNyb3NvZnQgQ29ycG9yYXRpb24xJjAkBgNVBAMTHU1pY3Jvc29mdCBU
# aW1lLVN0YW1wIFBDQSAyMDEwMB4XDTIzMTIwNjE4NDUzNFoXDTI1MDMwNTE4NDUz
# NFowgcsxCzAJBgNVBAYTAlVTMRMwEQYDVQQIEwpXYXNoaW5ndG9uMRAwDgYDVQQH
# EwdSZWRtb25kMR4wHAYDVQQKExVNaWNyb3NvZnQgQ29ycG9yYXRpb24xJTAjBgNV
# BAsTHE1pY3Jvc29mdCBBbWVyaWNhIE9wZXJhdGlvbnMxJzAlBgNVBAsTHm5TaGll
# bGQgVFNTIEVTTjpBMDAwLTA1RTAtRDk0NzElMCMGA1UEAxMcTWljcm9zb2Z0IFRp
# bWUtU3RhbXAgU2VydmljZTCCAiIwDQYJKoZIhvcNAQEBBQADggIPADCCAgoCggIB
# AMEVaCHaVuBXd4mnTWiqJoUG5hs1zuFIqaS28nXk2sH8MFuhSjDxY85M/FufuByY
# g4abAmR35PIXHso6fOvGegHeG6+/3V9m5S6AiwpOcC+DYFT+d83tnOf0qTWam4nb
# tLrFQMfih0WJfnUgJwqXoQbhzEqBwMCKeKFPzGuglZUBMvunxtt+fCxzWmKFmZy8
# i5gadvVNj22el0KFav0QBG4KjdOJEaMzYunimJPaUPmGd3dVoZN6k2rJqSmQIZXT
# 5wrxW78eQhl2/L7PkQveiNN0Usvm8n0gCiBZ/dcC7d3tKkVpqh6LHR7WrnkAP3hn
# AM/6LOotp2wFHe3OOrZF+sI0v5OaL+NqVG2j8npuHh8+EcROcMLvxPXJ9dRB0a2Y
# n+60j8A3GLsdXyAA/OJ31NiMw9tiobzLnHP6Aj9IXKP5oq0cdaYrMRc+21fMBx7E
# nUQfvBu6JWTewSs8r0wuDVdvqEzkchYDSMQBmEoTJ3mEfZcyJvNqRunazYQlBZqx
# BzgMxoXUSxDULOAKUNghgbqtSG518juTwv0ooIS59FsrmV1Fg0Cp12v/JIl+5m/c
# 9Lf6+0PpfqrUfhQ6aMMp2OhbeqzslExmYf1+QWQzNvphLOvp5fUuhibc+s7Ul5rj
# dJjOUHdPPzg6+5VJXs1yJ1W02qJl5ZalWN9q9H4mP8k5AgMBAAGjggFJMIIBRTAd
# BgNVHQ4EFgQUdJ4FrNZVzG7ipP07mNPYH6oB6uEwHwYDVR0jBBgwFoAUn6cVXQBe
# Yl2D9OXSZacbUzUZ6XIwXwYDVR0fBFgwVjBUoFKgUIZOaHR0cDovL3d3dy5taWNy
# b3NvZnQuY29tL3BraW9wcy9jcmwvTWljcm9zb2Z0JTIwVGltZS1TdGFtcCUyMFBD
# QSUyMDIwMTAoMSkuY3JsMGwGCCsGAQUFBwEBBGAwXjBcBggrBgEFBQcwAoZQaHR0
# cDovL3d3dy5taWNyb3NvZnQuY29tL3BraW9wcy9jZXJ0cy9NaWNyb3NvZnQlMjBU
# aW1lLVN0YW1wJTIwUENBJTIwMjAxMCgxKS5jcnQwDAYDVR0TAQH/BAIwADAWBgNV
# HSUBAf8EDDAKBggrBgEFBQcDCDAOBgNVHQ8BAf8EBAMCB4AwDQYJKoZIhvcNAQEL
# BQADggIBAIN03y+g93wL5VZk/f5bztz9Bt1tYrSw631niQQ5aeDsqaH5YPYuc8lM
# kogRrGeI5y33AyAnzJDLBHxYeAM69vCp2qwtRozg2t6u0joUj2uGOF5orE02cFnM
# dksPCWQv28IQN71FzR0ZJV3kGDcJaSdXe69Vq7XgXnkRJNYgE1pBL0KmjY6nPdxG
# ABhV9osUZsCs1xG9Ja9JRt4jYgOpHELjEFtGI1D7WodcMI+fSEaxd8v7KcNmdwJ+
# zM2uWBlPbheCG9PNgwdxeKgtVij/YeTKjDp0ju5QslsrEtfzAeGyLCuJcgMKeMtW
# wbQTltHzZCByx4SHFtTZ3VFUdxC2RQTtb3PFmpnr+M+ZqiNmBdA7fdePE4dhhVr8
# Fdwi67xIzM+OMABu6PBNrClrMsG/33stEHRk5s1yQljJBCkRNJ+U3fqNb7PtH+cb
# ImpFnce1nWVdbV/rMQIB4/713LqeZwKtVw6ptAdftmvxY9yCEckAAOWbkTE+HnGL
# W01GT6LoXZr1KlN5Cdlc/nTD4mhPEhJCru8GKPaeK0CxItpV4yqg+L41eVNQ1nY1
# 21sWvoiKv1kr259rPcXF+8Nmjfrm8s6jOZA579n6m7i9jnM+a02JUhxCcXLslk6J
# lUMjlsh3BBFqLaq4conqW1R2yLceM2eJ64TvZ9Ph5aHG2ac3kdgIMIIHcTCCBVmg
# AwIBAgITMwAAABXF52ueAptJmQAAAAAAFTANBgkqhkiG9w0BAQsFADCBiDELMAkG
# A1UEBhMCVVMxEzARBgNVBAgTCldhc2hpbmd0b24xEDAOBgNVBAcTB1JlZG1vbmQx
# HjAcBgNVBAoTFU1pY3Jvc29mdCBDb3Jwb3JhdGlvbjEyMDAGA1UEAxMpTWljcm9z
# b2Z0IFJvb3QgQ2VydGlmaWNhdGUgQXV0aG9yaXR5IDIwMTAwHhcNMjEwOTMwMTgy
# MjI1WhcNMzAwOTMwMTgzMjI1WjB8MQswCQYDVQQGEwJVUzETMBEGA1UECBMKV2Fz
# aGluZ3RvbjEQMA4GA1UEBxMHUmVkbW9uZDEeMBwGA1UEChMVTWljcm9zb2Z0IENv
# cnBvcmF0aW9uMSYwJAYDVQQDEx1NaWNyb3NvZnQgVGltZS1TdGFtcCBQQ0EgMjAx
# MDCCAiIwDQYJKoZIhvcNAQEBBQADggIPADCCAgoCggIBAOThpkzntHIhC3miy9ck
# eb0O1YLT/e6cBwfSqWxOdcjKNVf2AX9sSuDivbk+F2Az/1xPx2b3lVNxWuJ+Slr+
# uDZnhUYjDLWNE893MsAQGOhgfWpSg0S3po5GawcU88V29YZQ3MFEyHFcUTE3oAo4
# bo3t1w/YJlN8OWECesSq/XJprx2rrPY2vjUmZNqYO7oaezOtgFt+jBAcnVL+tuhi
# JdxqD89d9P6OU8/W7IVWTe/dvI2k45GPsjksUZzpcGkNyjYtcI4xyDUoveO0hyTD
# 4MmPfrVUj9z6BVWYbWg7mka97aSueik3rMvrg0XnRm7KMtXAhjBcTyziYrLNueKN
# iOSWrAFKu75xqRdbZ2De+JKRHh09/SDPc31BmkZ1zcRfNN0Sidb9pSB9fvzZnkXf
# tnIv231fgLrbqn427DZM9ituqBJR6L8FA6PRc6ZNN3SUHDSCD/AQ8rdHGO2n6Jl8
# P0zbr17C89XYcz1DTsEzOUyOArxCaC4Q6oRRRuLRvWoYWmEBc8pnol7XKHYC4jMY
# ctenIPDC+hIK12NvDMk2ZItboKaDIV1fMHSRlJTYuVD5C4lh8zYGNRiER9vcG9H9
# stQcxWv2XFJRXRLbJbqvUAV6bMURHXLvjflSxIUXk8A8FdsaN8cIFRg/eKtFtvUe
# h17aj54WcmnGrnu3tz5q4i6tAgMBAAGjggHdMIIB2TASBgkrBgEEAYI3FQEEBQID
# AQABMCMGCSsGAQQBgjcVAgQWBBQqp1L+ZMSavoKRPEY1Kc8Q/y8E7jAdBgNVHQ4E
# FgQUn6cVXQBeYl2D9OXSZacbUzUZ6XIwXAYDVR0gBFUwUzBRBgwrBgEEAYI3TIN9
# AQEwQTA/BggrBgEFBQcCARYzaHR0cDovL3d3dy5taWNyb3NvZnQuY29tL3BraW9w
# cy9Eb2NzL1JlcG9zaXRvcnkuaHRtMBMGA1UdJQQMMAoGCCsGAQUFBwMIMBkGCSsG
# AQQBgjcUAgQMHgoAUwB1AGIAQwBBMAsGA1UdDwQEAwIBhjAPBgNVHRMBAf8EBTAD
# AQH/MB8GA1UdIwQYMBaAFNX2VsuP6KJcYmjRPZSQW9fOmhjEMFYGA1UdHwRPME0w
# S6BJoEeGRWh0dHA6Ly9jcmwubWljcm9zb2Z0LmNvbS9wa2kvY3JsL3Byb2R1Y3Rz
# L01pY1Jvb0NlckF1dF8yMDEwLTA2LTIzLmNybDBaBggrBgEFBQcBAQROMEwwSgYI
# KwYBBQUHMAKGPmh0dHA6Ly93d3cubWljcm9zb2Z0LmNvbS9wa2kvY2VydHMvTWlj
# Um9vQ2VyQXV0XzIwMTAtMDYtMjMuY3J0MA0GCSqGSIb3DQEBCwUAA4ICAQCdVX38
# Kq3hLB9nATEkW+Geckv8qW/qXBS2Pk5HZHixBpOXPTEztTnXwnE2P9pkbHzQdTlt
# uw8x5MKP+2zRoZQYIu7pZmc6U03dmLq2HnjYNi6cqYJWAAOwBb6J6Gngugnue99q
# b74py27YP0h1AdkY3m2CDPVtI1TkeFN1JFe53Z/zjj3G82jfZfakVqr3lbYoVSfQ
# JL1AoL8ZthISEV09J+BAljis9/kpicO8F7BUhUKz/AyeixmJ5/ALaoHCgRlCGVJ1
# ijbCHcNhcy4sa3tuPywJeBTpkbKpW99Jo3QMvOyRgNI95ko+ZjtPu4b6MhrZlvSP
# 9pEB9s7GdP32THJvEKt1MMU0sHrYUP4KWN1APMdUbZ1jdEgssU5HLcEUBHG/ZPkk
# vnNtyo4JvbMBV0lUZNlz138eW0QBjloZkWsNn6Qo3GcZKCS6OEuabvshVGtqRRFH
# qfG3rsjoiV5PndLQTHa1V1QJsWkBRH58oWFsc/4Ku+xBZj1p/cvBQUl+fpO+y/g7
# 5LcVv7TOPqUxUYS8vwLBgqJ7Fx0ViY1w/ue10CgaiQuPNtq6TPmb/wrpNPgkNWcr
# 4A245oyZ1uEi6vAnQj0llOZ0dFtq0Z4+7X6gMTN9vMvpe784cETRkPHIqzqKOghi
# f9lwY1NNje6CbaUFEMFxBmoQtB1VM1izoXBm8qGCA00wggI1AgEBMIH5oYHRpIHO
# MIHLMQswCQYDVQQGEwJVUzETMBEGA1UECBMKV2FzaGluZ3RvbjEQMA4GA1UEBxMH
# UmVkbW9uZDEeMBwGA1UEChMVTWljcm9zb2Z0IENvcnBvcmF0aW9uMSUwIwYDVQQL
# ExxNaWNyb3NvZnQgQW1lcmljYSBPcGVyYXRpb25zMScwJQYDVQQLEx5uU2hpZWxk
# IFRTUyBFU046QTAwMC0wNUUwLUQ5NDcxJTAjBgNVBAMTHE1pY3Jvc29mdCBUaW1l
# LVN0YW1wIFNlcnZpY2WiIwoBATAHBgUrDgMCGgMVAIAGiXW7XDDBiBS1SjAyepi9
# u6XeoIGDMIGApH4wfDELMAkGA1UEBhMCVVMxEzARBgNVBAgTCldhc2hpbmd0b24x
# EDAOBgNVBAcTB1JlZG1vbmQxHjAcBgNVBAoTFU1pY3Jvc29mdCBDb3Jwb3JhdGlv
# bjEmMCQGA1UEAxMdTWljcm9zb2Z0IFRpbWUtU3RhbXAgUENBIDIwMTAwDQYJKoZI
# hvcNAQELBQACBQDqSRZ5MCIYDzIwMjQwNzIyMTczMzQ1WhgPMjAyNDA3MjMxNzMz
# NDVaMHQwOgYKKwYBBAGEWQoEATEsMCowCgIFAOpJFnkCAQAwBwIBAAICEYQwBwIB
# AAICE7EwCgIFAOpKZ/kCAQAwNgYKKwYBBAGEWQoEAjEoMCYwDAYKKwYBBAGEWQoD
# AqAKMAgCAQACAwehIKEKMAgCAQACAwGGoDANBgkqhkiG9w0BAQsFAAOCAQEAHOho
# z6pVQId0e0IdIHWl1kC+4jhxAFvT2LyC3GT7KcFokVmxUW1rxc2QjaZjAxXrYyEX
# 7M/JSU3XkKmWi4cdO+QRiiZC7kz2T/3knG3TbbcOXwWu4rQ2efOVmHSW4YH2GBu+
# y6Rvv7u2uswRu7t346fRih5FGwMkFTWNDTGZQP8HPWMMXweyBY5nNePmdtkdnJJA
# hLVK32u2UZVB47KB0RFBybND764qGFaS/5aJH7oPgEztTnQX+ZA9za6W+jn+/SNj
# u2dFWtiDsgpLy50brMdidZvobll0DQOWCkA5VtzQTS9ohaqTxZ4vSB4jBoMMFd+0
# gtTL5nAYlE+DahRipjGCBA0wggQJAgEBMIGTMHwxCzAJBgNVBAYTAlVTMRMwEQYD
# VQQIEwpXYXNoaW5ndG9uMRAwDgYDVQQHEwdSZWRtb25kMR4wHAYDVQQKExVNaWNy
# b3NvZnQgQ29ycG9yYXRpb24xJjAkBgNVBAMTHU1pY3Jvc29mdCBUaW1lLVN0YW1w
# IFBDQSAyMDEwAhMzAAAB6+AYbLW27zjtAAEAAAHrMA0GCWCGSAFlAwQCAQUAoIIB
# SjAaBgkqhkiG9w0BCQMxDQYLKoZIhvcNAQkQAQQwLwYJKoZIhvcNAQkEMSIEIK47
# g4GGjY4pj3OchdktrSG7C54ivohZqYfxB49uQArZMIH6BgsqhkiG9w0BCRACLzGB
# 6jCB5zCB5DCBvQQgzrdrvl9pA+F/xIENO2TYNJSAht2LNezPvqHUl2EsLPgwgZgw
# gYCkfjB8MQswCQYDVQQGEwJVUzETMBEGA1UECBMKV2FzaGluZ3RvbjEQMA4GA1UE
# BxMHUmVkbW9uZDEeMBwGA1UEChMVTWljcm9zb2Z0IENvcnBvcmF0aW9uMSYwJAYD
# VQQDEx1NaWNyb3NvZnQgVGltZS1TdGFtcCBQQ0EgMjAxMAITMwAAAevgGGy1tu84
# 7QABAAAB6zAiBCCzA7gOCCtaufHy6EU0bB0RmxXUe8ajI6lY/YE/3zHbOzANBgkq
# hkiG9w0BAQsFAASCAgCbdxOtwG6x3B92Ir+Wn8lkq4ZA1jkUA61AGiok+n/UDyd4
# ohGshxYACaEeNPk4vb4p+X0k+nKbm/NnhB3MPKkSKddcnkAujlcmH1oQiWJNLYqR
# 0YtbE4c9y33eRxrpH3FxrKQSM2E/yFHjRz9j7GXYzU1YYOS4wEPH9bkIfcMvbAva
# K7tDVpquN3LONsLskTb+QFxdrWxPFPrMM/E7gAtOJ28BuH3l6kCbh3d5/FRVfbE1
# xjwTzLqK+6LJh/sQ3C7P5R6KtMzyYOFlwGA35dqHHedRCASpyzwcmS4ret5kwRpk
# UFGlNzBHrau2ye3KK06fTXVNyS5obtP0WpHx/HUsdP0EFRxNmwXqL2g+Uoxd1OPU
# 7WnsAU8OF4B9ogC4tk7zjbL8DeJqDJm34eIArkgw4gCfRFXQkYnSBZr+PipKVWBq
# fyW9NI7fnwJIFEUHmA0TZvM30jcqApeJyuiYjF5z4wYehggid1dRQXUAFadq0k1F
# OxG/g/1BBzanK0ZPcdbSfeOSEGrlMdQKM1bFceHeRKBxutTBg8q7u12Iwc/eLo+L
# yreiEaDki3FVhmS6Fj2rbEla1TfqorT/Ob/eScIxDUgZXVjGEsIHz4KQKCSYfGPe
# cO7VXOBREH8yUnLC+5N3frpH6vYR7PqiF8RolQkfcGNJeCNpFvCUnQhODd+e0Q==
# SIG # End signature block
